/*
 *  tclgdbm.c --
 *
 *   This file contains a simple Tcl-Wrapper for gdbm-1.8.0.
 *   Copyright (C) 2000, Stefan Vogel
 */
/*  TCLGDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    TCLGDBM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCLGDBM; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  stefan_vogel@usa.net

*************************************************************************/

#if defined(__WIN32__)
#   if defined(_MSC_VER)
#     define EXPORT(a,b) __declspec(dllexport) a b
#   else
#     if defined(__BORLANDC__)
#         define EXPORT(a,b) a _export b
#     else
#         define EXPORT(a,b) a b
#     endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <tcl.h>

#include <gdbm.h>

extern char* gdbm_version;

#define TCLGDBM_VERSION "0.3"

/* -------------------- global variables -------------------- */
/* handle to opened gdbm-file, used in Gdbm_Cmd */
Tcl_HashTable thtbl;

/* pointer to the tcl-string (if provided in gdbm_open call),
   the tcl-function specified with this string is called
	 in fatal-error-case (hopefully never) */
char* gFatalFunc = NULL;

/* Tcl-Interpreter for fatal-function */
Tcl_Interp* gInterp = NULL;

/*
 * Prototypes for procedures defined later in this file:
 */
int Gdbm_Open_Cmd _ANSI_ARGS_((ClientData, Tcl_Interp*, int, char**));
int Gdbm_Cmd _ANSI_ARGS_((ClientData, Tcl_Interp*, int, char**));

/*
 * some defines to make checking of arguments and returning
 * and setting Errors easier
 */
#define CHECK_ARGS(ARGC, MIN, MAX, USG) if ((ARGC<MIN+1)||(ARGC>MAX+1)) {Tcl_AppendResult(interp, "Usage: ", USG, NULL);return (TCL_ERROR);}

char sErrno[5];

#define SET_ERROR(errno) sprintf(sErrno, "%d", errno);Tcl_SetVar(interp, "GDBM_ERRNO", sErrno, 0);return TCL_ERROR

#define SET_OK Tcl_SetVar(interp, "GDBM_ERRNO", "0", 0);return TCL_OK



/*----------------------------------------------------------------------
 * GetIndexFromArg -- search for a specified string in given string-table
 * 
 * Because of performance no Tcl-Obj-Command was used here. This has the
 * affect that no Obj-Command-Utility-Functions could be used, too.
 * This one is the only Utility-Function needed, so it is reimplemented
 * here.
 *
 * Results: 
 *  position/index of string in string-table
 *  -1 : string not found in string-table
 *----------------------------------------------------------------------
 */
int 
GetIndexFromArg(arg, tablePtr)
     char* arg;
     char** tablePtr;
{
  int i;
  char **entryPtr;

  for (entryPtr = tablePtr, i = 0; *entryPtr != NULL;
       entryPtr = (char**) ((long) entryPtr + sizeof(char*)), i++) {
    if (!strcmp(*entryPtr, arg)) return i;
  }
  return -1;
}


/*----------------------------------------------------------------------
 * isNumber -- check a given string if it is a valid positive integer
 * Results: 
 *  1 : string contains only digits
 *  0 : string is no positive integer
 *----------------------------------------------------------------------
 */
int
isNumber(string)
     char* string;
{
  unsigned int i;

  for (i = 0; i < strlen(string); i++) {
    if (!isdigit((int)string[i])) return 0;
  }
  return 1;
}


/*----------------------------------------------------------------------
 * gdbmFatalFunc -- standard function called from gdbm in fatal error
 *
 *  If a Tcl-Function has been provided with -fatal_func in call to
 *  gdbm_open, this function will be called here.
 *  Hopefully this should be never.
 *----------------------------------------------------------------------
 */
void
gdbmFatalFunc(message)
    char* message;
{
  char* script;

  if (gFatalFunc && gInterp) {
    script = malloc(strlen(gFatalFunc)+strlen(message)+5);
    sprintf(script, "%s {%s}", gFatalFunc, message);
    Tcl_Eval(gInterp, script);
    free(script);
  }
	exit(1);
}


/*----------------------------------------------------------------------
 * Gdbm_Error_Cmd -- Return a textual error-description for error-number
 *
 * This is the aequivalent for gdbm_strerror. If given an errorcode
 * the value of the global Tcl-Variable GDBM_ERRNO an english error-message
 * is returned.
 *
 * Results:
 *  Error-message.
 *
 * Side effects:
 *  None.
 *----------------------------------------------------------------------
 */
int
Gdbm_Error_Cmd(errno, interp, argc, argv)
    ClientData errno;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
  CHECK_ARGS(argc, 1, 1, "gdbm_strerror $GDBM_ERRNO");
  Tcl_SetResult(interp, gdbm_strerror(atoi(argv[1])), TCL_STATIC);
  SET_OK;
}


/*----------------------------------------------------------------------
 * Gdbm_Open_Cmd -- Open/Create a GDBM-File
 *
 *  This procedure is invoked to open/create a GDBM-Databasefile
 *  It expects several parameters:
 *    -writer, 
 *    -reader:     the calling process does only read from file or write
 *    -wrcreat:    create the db-file, if it is not existent
 *    -newdb:      create the db-file, regardless if one exists
 *    -nolock:     do not lock the db-file (should be used with care)
 *    -sync:       db-operations are synchronized to the disk
 *    -block_size: see gdbm-manual
 *    -mode:       file-mode, if the db-file is created (directly
 *                 passed to open(2))
 *    -fatal_func: a tcl-function-name (which gets one argument)
 *    and the name of the gdbm-file
 *
 *  The parameters: -writer, -reader, -wrcreat, -newdb, -nolock
 *  represent the following gdbm-defines:
 *  GDBM_WRITER, GDBM_READER, GDBM_WRCREAT, GDBM_NEWDB
 *  For db-writers the following flags could be specified:
 *  GDBM_NOLOCK or GDBM_SYNC.
 *
 * Results:
 *  A unique gdbm-file-descriptor is returned. This is used for further
 *  access to the gdbm-file as a Tcl-Command.
 *
 * Side effects:
 *  A Tcl-Command is created.
 *----------------------------------------------------------------------
 */
int
Gdbm_Open_Cmd(notused, interp, argc, argv)
    ClientData notused;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
  int i, index, flags, block_size, mode, already_created;
	GDBM_FILE hd_gdbm = NULL;
	Tcl_HashEntry *thetr;
  static int db_number = 0;
  char gdbm_name[20];
  static char usage[] = "gdbm_open ?options? filename\noptions=-reader, -writer, -wrcreat, -newdb, -nolock -sync\n        -block_size <size> -mode <filemode> -fatal_func <command>\n  where <size>     512, ...\n        <filemode> accessmode of file, e.g. [expr 0664]\n        <command>  tcl-proc with one parameter called in case of fatal error\n        for further information see gdbm-man-page";
  static char *switches[] = 
  {"-reader", "-writer", "-wrcreat", "-newdb", "-nolock", "-sync", 
   "-block_size", "-mode", "-fatal_func", (char*) NULL};
  static int gdbm_flags[] = 
  { GDBM_READER, GDBM_WRITER, GDBM_WRCREAT, GDBM_NEWDB, GDBM_NOLOCK, 
    GDBM_SYNC };

  flags = 0;
  block_size = 0;
  mode = 0644;

  if (argc < 2) {
    Tcl_AppendResult(interp, usage, NULL);
    return TCL_ERROR;
  }

  /*
   * Parse arguments
   */
  for (i = 1; i < argc-1; i++) {
    index = GetIndexFromArg(argv[i], switches);
    switch (index) {
    case 0: /* -reader */
    case 1: /* -writer */
    case 2: /* -wrcreat */
    case 3: /* -newdb */
    case 4: /* -nolock */
    case 5: /* -sync */
      flags = flags | gdbm_flags[index];
      break;

    case 6: /* -block_size */
      if ((i++ < argc-1) && isNumber(argv[i])) {
        block_size = atoi(argv[i]);
      }
      else {
        Tcl_AppendResult(interp, "-block_size numbervalue\n\n", usage, NULL);
        return TCL_ERROR;
      }
      break;

    case 7: /* -mode */
      if ((i++ < argc-1) && isNumber(argv[i])) {
        mode = atoi(argv[i]);
      }
      else {
        Tcl_AppendResult(interp, "-mode numbervalue\n\n", usage, NULL);
        return TCL_ERROR;
      }
      break;

    case 8: /* -fatal_func */
      if (i++ < argc-1) {
        free(gFatalFunc);
        gFatalFunc = malloc(strlen(argv[i])+1);
        strcpy(gFatalFunc, argv[i]);
        gInterp = interp;
      }
      break;

    default: /* error */
      Tcl_AppendResult(interp, "error: gdbm_open - No such argument: ",
                       argv[i], "\n", usage, NULL);
      SET_ERROR(GDBM_OPT_ILLEGAL);
    }
  }

  /* call to gdbm_open with the collected parameters */
  hd_gdbm = gdbm_open(argv[argc-1], block_size, flags, mode, gdbmFatalFunc);
  if (hd_gdbm == NULL) {
    Tcl_AppendResult(interp, "error: gdbm_open - ", 
                     gdbm_strerror(gdbm_errno), "\n", 
                     NULL);
    SET_ERROR(gdbm_errno);
  }

  /* create the unique gdbm-descriptor */
  sprintf(gdbm_name, "gdbm%d", db_number);
  db_number++;

	/* put gdbm-file-handle in HashTable: */
	thetr = Tcl_CreateHashEntry(&thtbl, gdbm_name, &already_created);
	if (thetr == NULL) gdbmFatalFunc("Could not create Tcl-Hashentry!");
	Tcl_SetHashValue(thetr, hd_gdbm);

  /* create the tcl-command with this descriptor, pass the gdbm-handle */
  Tcl_CreateCommand(interp, gdbm_name, Gdbm_Cmd, 
                    (ClientData) hd_gdbm, NULL);

  /* Make a copy of this name */
  Tcl_SetResult(interp, gdbm_name, TCL_VOLATILE);
  SET_OK;
}


/*----------------------------------------------------------------------
 * Gdbm_Cmd -- Implementation of all related gdbm-commands
 *
 * The command created with Gdbm_Open_Cmd has several options:
 * (<gdbm0> is the command created with Gdbm_Open_Cmd)
 *   close:       (e.g.:  gdbm0 close)
 *                close the associated gdbm-file
 *   fetch:       (e.g.:  gdbm0 fetch key)
 *                returns the value stored in gdbm-file with key "key"
 *   delete:      (e.g.:  gdbm0 delete key)
 *                deletes the key-value-pair with key "key"
 *   store:       (either:  gdbm0 -insert store key value
 *                     or:  gdbm0 -replace store key value)
 *                Inserts/Replaces the given key-value-pair
 *                If key already exists and -insert is specified an error
 *                is thrown. -replace could be called without error even
 *                if no key "key" is existent.
 *   exists:      (e.g.: gdbm0 exists key)
 *                checks if given key already exists: returns 1 if it exists
 *                0 if it doesn't
 *   firstkey:    (e.g.: gdbm0 firstkey)
 *                returns the first key. Used for iteration through the
 *                gdbm-file. 
 *   nextkey:     (e.g.: gdbm0 nextkey key)
 *                return the nextkey to "key", which is retrieved with
 *                first- or nextkey
 *   reorganize:  (e.g.: gdbm0 reorganize)
 *                when you had deleted lots of entries, this would shrink
 *                the database-file
 *
 *  These commands all directy match to corresponding gdbm-commands (see
 *  gdbm(3)).
 *  There are more commands, which do not correspond to any gdbm-commands,
 *  but are provided for simplicity:
 * 
 *   count:       (e.g.: gdbm0 count)
 *                returns number of entries in gdbm-file
 *   maxkey:      (e.g.: gdbm0 maxkey)
 *                if you had only integer-keys and you need to create
 *                a unique new-key, this one searches for the maximal
 *                existing key and returns it.
 *
 *  In case of an error an Tclerror is returned with an errormessage and
 *  the global gdbm_errno-variable is provided in Tcl-Variable GDBM_ERRNO.
 * 
 * Results:
 *  Depends on option to command, see description above.
 *----------------------------------------------------------------------
 */
int
Gdbm_Cmd(cd_gdbm, interp, argc, argv)
    ClientData cd_gdbm;
    Tcl_Interp* interp;
    int argc;
    char* argv[];
{
	Tcl_HashEntry *thetr;
  datum key, content, result;
  int insert_replace_flag, count;
  char number[13];
	GDBM_FILE hd_gdbm;
  static char *switches[] = 
  {"close", "fetch", "delete", "firstkey", "nextkey", "reorganize", "exists", 
   "-insert", "-replace", "count", "maxkey", (char*) NULL};
  static char usage[] = "gdbmN option ?arg arg ...?\nwhere option:\nclose, fetch, delete, [-insert|-replace] store, exists, firstkey, nextkey,\nreorganize, count or maxkey";

  /* retrieve the gdbm-descriptor from clientdata */
  GDBM_FILE gdbm = (GDBM_FILE) cd_gdbm;

  /* set store-flag per default to REPLACE */
  insert_replace_flag = GDBM_REPLACE;

  if (argc < 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], usage, NULL);
    return TCL_ERROR;
  }

  switch(GetIndexFromArg(argv[1], switches)) {
  case 0: /* ----- close ----- */
    CHECK_ARGS(argc, 1, 1, "gdbmN close");
    if (gdbm != NULL) gdbm_close(gdbm);
    Tcl_DeleteCommand(interp, argv[0]);

		/* now that the command is deleted we free memory,
		   gdbm is freed in gdbm_close */
		thetr = Tcl_FindHashEntry(&thtbl, argv[0]);
		if (thetr != NULL) Tcl_DeleteHashEntry(thetr);

    Tcl_SetResult(interp, "", TCL_STATIC);
    SET_OK;
    break;

  case 1: /* -----fetch ----- */
    CHECK_ARGS(argc, 2, 2, "gdbmN fetch key");
    key.dptr = argv[2];
    key.dsize = strlen(key.dptr)+1;
    result = gdbm_fetch(gdbm, key);
    if (result.dptr == NULL) {
      Tcl_AppendResult(interp, "error: gdbm_fetch - ",
                       gdbm_strerror(gdbm_errno), "\n",
                       NULL);
      SET_ERROR(gdbm_errno);
    } 
    Tcl_SetResult(interp, result.dptr, TCL_VOLATILE);
    free(result.dptr);
    SET_OK;
    break;

  case 2: /* ----- delete ----- */
    CHECK_ARGS(argc, 2, 2, "gdbmN delete key");
    key.dptr = argv[2];
    key.dsize = strlen(key.dptr) +1;
    
    /* retcode == -1 --> no such key or requester is reader
     *         ==  0 --> successful delete
     */
    if (gdbm_delete(gdbm, key) != 0) {
      Tcl_AppendResult(interp, "error: gdbmN delete - ",
                       gdbm_strerror(gdbm_errno),
                       NULL);
      SET_ERROR(gdbm_errno);
    }
    SET_OK;
    break;

  case 3: /* ----- firstkey ----- */
    CHECK_ARGS(argc, 1, 1, "gdbmN firstkey");
    result = gdbm_firstkey(gdbm);
    if (result.dptr == NULL) {
      /* no result found return an empty string */
      Tcl_SetResult(interp, "", TCL_STATIC);
    } else {
      Tcl_SetResult(interp, result.dptr, TCL_VOLATILE);
      free(result.dptr);
    }
    SET_OK;
    break;

  case 4: /* ----- nextkey ----- */
    CHECK_ARGS(argc, 2, 2, "gdbmN nextkey key");
    key.dptr = argv[2];
    key.dsize = strlen(key.dptr) +1;
    result = gdbm_nextkey(gdbm, key);
    if (result.dptr == NULL) {
      /* no result found return an empty string */
      Tcl_SetResult(interp, "", TCL_STATIC);
    } else {
      Tcl_SetResult(interp, result.dptr, TCL_VOLATILE);
      free(result.dptr);
    }
    SET_OK;
    break;

  case 5: /* ----- reorganize ----- */
    CHECK_ARGS(argc, 1, 1, "gdbmN reorganize");
    if (gdbm_reorganize(gdbm)) {
      Tcl_AppendResult(interp, "error: gdbmN reorganize",
                       gdbm_strerror(gdbm_errno),
                       NULL);
      SET_ERROR(gdbm_errno);
    } else {
			/* get according gdbm-file-handle from hash-table */
			thetr = Tcl_FindHashEntry(&thtbl, argv[0]);
			if (thetr == NULL) gdbmFatalFunc("Could not find Tcl-Hashentry!");
			hd_gdbm =  (GDBM_FILE) Tcl_GetHashValue(thetr);
			/* the gdbm-structure has changed we had to create the
         command again, to let tcl have the new settings
         (especially the new gdbm-handle for clientdata)
         Tcl_DeleteCommand(interp, argv[0]) must not be called,
         this is done by Tcl_CreateCommand */
      Tcl_CreateCommand(interp, argv[0], Gdbm_Cmd, 
                    (ClientData) hd_gdbm, NULL);
      Tcl_SetResult(interp, "", TCL_STATIC);
      SET_OK;
    }
    break;

  case 6: /* ----- exists ----- */
    CHECK_ARGS(argc, 2, 2, "gdbmN exists key");
    key.dptr = argv[2];
    key.dsize = strlen(key.dptr) +1;
    if (gdbm_exists(gdbm, key))
      Tcl_SetResult(interp, "1", TCL_STATIC);
    else
      Tcl_SetResult(interp, "0", TCL_STATIC);
    SET_OK;
    break;

  case 7: /* ----- -insert -----   trickle through */
    insert_replace_flag = GDBM_INSERT;
  case 8: /* ----- -replace ----- */
    CHECK_ARGS(argc, 4, 4, "gdbmN [-insert|-replace] store key data");
    if (strcmp("store", argv[2])) {
      Tcl_AppendResult(interp, "error gdbmN [-insert|-replace] store key data",
                       NULL);
      SET_ERROR(GDBM_OPT_ILLEGAL);
    }
    key.dptr = argv[3];
    key.dsize = strlen(key.dptr) +1;
    content.dptr = argv[4];
    content.dsize = strlen(content.dptr) +1;

    /* insert-replace-flag is one of GDBM_INSERT==0 or GDBM_REPLACE==1
     */
    if (gdbm_store(gdbm, key, content, insert_replace_flag) != 0) {
      Tcl_AppendResult(interp, "error: Item not inserted/replaced", NULL);
      SET_ERROR(gdbm_errno);
    }
    SET_OK;
    break;

  case 9: /* ----- count ----- */   /* extension to basics of gdbm */
    CHECK_ARGS(argc, 1, 1, "gdbmN count");
    count = 0;
    result = gdbm_firstkey (gdbm);
    while (result.dptr != NULL) {
      count++;
      key = result;
      result = gdbm_nextkey (gdbm, key);
      free(key.dptr);
    }
    sprintf(number, "%d", count);
    Tcl_SetResult(interp, number, TCL_STATIC);
    SET_OK;
    break;

  case 10: /* ----- maxkey ----- */   /* extension to basics of gdbm */
    CHECK_ARGS(argc, 1, 1, "gdbmN maxkey");
    count = 0;
    result = gdbm_firstkey (gdbm);
    while (result.dptr != NULL) {
      if (isNumber(result.dptr) && (atoi(result.dptr) > count)) {
        count = atoi(result.dptr);
      }
      key = result;
      result = gdbm_nextkey (gdbm, key);
      free(key.dptr);
    }
    sprintf(number, "%d", count);
    Tcl_SetResult(interp, number, TCL_STATIC);
    SET_OK;
    break;
    /* ----- while ----- extension to basics of gdbm
       seems to be slower than tcl-only gdbm_while (so this is commented)
       o remember to get the next key, before calling Tcl_Eval
         because in case of a delete in Tcl_Eval, we already got
         the next key for the loop
  case 11:
    CHECK_ARGS(argc, 4, 4, "gdbmN while key value body");
    ret_value = TCL_OK;
    key = gdbm_firstkey (gdbm);
    while (key.dptr != NULL) {
      result = gdbm_fetch(gdbm, key);
      backup = key;
      Tcl_SetVar(interp, argv[2], backup.dptr, 0);
      Tcl_SetVar(interp, argv[3], result.dptr, 0);
      free(result.dptr);

      key = gdbm_nextkey (gdbm, key);

      ret_value = Tcl_Eval(interp, argv[4]);
      if (ret_value == TCL_CONTINUE) continue;
      if (ret_value == TCL_BREAK) break;
      if (ret_value != TCL_OK) return TCL_ERROR;
    }
    return TCL_OK
    break;
    */
  default: /* ----- error ----- */
    Tcl_AppendResult(interp, "No such option: ", argv[1], "\n", usage, NULL);
    SET_ERROR(GDBM_OPT_ILLEGAL);
  }
}


/*----------------------------------------------------------------------
 * Tclgdbm_Init -- Initialization of Tclgdbm
 *
 * Create the gdbm_open and gdbm_strerror-command and provide
 * the used GDBM_VERSION in Tcl-Variable GDBM_VERSION
 *
 * #define DllEntryPoint DllMain 
 *----------------------------------------------------------------------
 */
EXPORT(int, Tclgdbm_Init)(interp)
    Tcl_Interp* interp;
{
	int i;
	char sErrno[5], sErrMsg[128];
	int iErrors[] = {	
		GDBM_NO_ERROR, GDBM_MALLOC_ERROR, GDBM_BLOCK_SIZE_ERROR, 
		GDBM_FILE_OPEN_ERROR, GDBM_FILE_WRITE_ERROR, GDBM_FILE_SEEK_ERROR, 
		GDBM_FILE_READ_ERROR, GDBM_BAD_MAGIC_NUMBER, GDBM_EMPTY_DATABASE, 
		GDBM_CANT_BE_READER, GDBM_CANT_BE_WRITER, GDBM_READER_CANT_DELETE, 
		GDBM_READER_CANT_STORE, GDBM_READER_CANT_REORGANIZE, GDBM_UNKNOWN_UPDATE,
		GDBM_ITEM_NOT_FOUND, GDBM_REORGANIZE_FAILED, GDBM_CANNOT_REPLACE, 
		GDBM_ILLEGAL_DATA, GDBM_OPT_ALREADY_SET, GDBM_OPT_ILLEGAL};
	char *sErrors[] = {
		"GDBM_NO_ERROR","GDBM_MALLOC_ERROR","GDBM_BLOCK_SIZE_ERROR", 
		"GDBM_FILE_OPEN_ERROR","GDBM_FILE_WRITE_ERROR","GDBM_FILE_SEEK_ERROR", 
		"GDBM_FILE_READ_ERROR","GDBM_BAD_MAGIC_NUMBER","GDBM_EMPTY_DATABASE", 
		"GDBM_CANT_BE_READER","GDBM_CANT_BE_WRITER","GDBM_READER_CANT_DELETE", 
		"GDBM_READER_CANT_STORE","GDBM_READER_CANT_REORGANIZE","GDBM_UNKNOWN_UPDATE",
		"GDBM_ITEM_NOT_FOUND","GDBM_REORGANIZE_FAILED","GDBM_CANNOT_REPLACE",
		"GDBM_ILLEGAL_DATA","GDBM_OPT_ALREADY_SET","GDBM_OPT_ILLEGAL"};

	/* initialize hash-table for gdbm-file-descriptors */
	Tcl_InitHashTable(&thtbl, TCL_STRING_KEYS);

  Tcl_SetVar(interp, "GDBM_VERSION", gdbm_version, 0);
	for (i = 0; i < sizeof(iErrors)/sizeof(int); i++) {
		sprintf(sErrno, "%d", iErrors[i]);
		sprintf(sErrMsg, "gdbm_error(%s)", sErrors[i]);
		Tcl_SetVar(interp, sErrMsg, sErrno, 0);
	}
  Tcl_CreateCommand(interp, "gdbm_open", Gdbm_Open_Cmd, NULL, NULL);
  Tcl_CreateCommand(interp, "gdbm_strerror", Gdbm_Error_Cmd, NULL, NULL);
	
	Tcl_PkgProvide(interp, "gdbm", TCLGDBM_VERSION);

  return TCL_OK;
}
